# rxscrite/parser.py

from .tokens import (
    TOKEN_INTEGER, TOKEN_FLOAT, TOKEN_PLUS, TOKEN_MINUS, TOKEN_MULTIPLY,
    TOKEN_DIVIDE, TOKEN_POWER, TOKEN_LPAREN, TOKEN_RPAREN, TOKEN_EOF,
    TOKEN_IDENTIFIER, TOKEN_KEYWORD, TOKEN_ASSIGN, TOKEN_COMMA, TOKEN_NEWLINE,
    TOKEN_STRING
)
from .nodes import (
    NumberNode, BinaryOpNode, UnaryOpNode, VariableAssignNode,
    VariableAccessNode, FunctionCallNode, ProgramNode, IdentifierNode,
    StringNode, TruthNode, FalsityNode, EmptyNode
)
from .errors import RxParserError

class Parser:
    def __init__(self, tokens, filename="<stdin>"):
        self.tokens = tokens
        self.filename = filename
        self.token_idx = -1
        self.current_token = None
        self.advance()

    def advance(self):
        self.token_idx += 1
        if self.token_idx < len(self.tokens):
            self.current_token = self.tokens[self.token_idx]
        else:
            self.current_token = self.tokens[-1] if self.tokens and self.tokens[-1].type == TOKEN_EOF else None
        return self.current_token

    def peek(self):
        if self.token_idx + 1 < len(self.tokens):
            return self.tokens[self.token_idx + 1]
        return None

    def error(self, expected_token_type=None, message=""):
        if not message:
            if expected_token_type and self.current_token:
                message = f"Expected {expected_token_type}, but got {self.current_token.type}"
            elif self.current_token:
                message = f"Unexpected token: {self.current_token}"
            else:
                message = "Unexpected end of file."
        
        line = self.current_token.line if self.current_token else (self.tokens[-1].line if self.tokens else None)
        column = self.current_token.column if self.current_token else (self.tokens[-1].column if self.tokens else None)
        raise RxParserError(message, line=line, column=column)

    def eat(self, token_type, token_value=None):
        if self.current_token is None:
            self.error(expected_token_type, f"Unexpected EOF. Expected {token_type}")
        
        if self.current_token.type == token_type:
            if token_value is not None and self.current_token.value != token_value:
                self.error(expected_token_type, f"Expected token value '{token_value}' but got '{self.current_token.value}'")
            eaten_token = self.current_token
            self.advance()
            return eaten_token
        else:
            self.error(expected_token_type)

    def program(self):
        prog_node = ProgramNode()
        while self.current_token and self.current_token.type == TOKEN_NEWLINE:
            self.eat(TOKEN_NEWLINE)

        while self.current_token and self.current_token.type != TOKEN_EOF:
            if self.current_token.type == TOKEN_NEWLINE:
                self.eat(TOKEN_NEWLINE)
                continue
            
            stmt = self.statement()
            if stmt:
                prog_node.statements.append(stmt)

            if self.current_token.type != TOKEN_EOF:
                if self.current_token.type == TOKEN_NEWLINE:
                    while self.current_token and self.current_token.type == TOKEN_NEWLINE:
                        self.eat(TOKEN_NEWLINE)
                else:
                    self.error(TOKEN_NEWLINE, "Expected newline or EOF after statement.")
            
            while self.current_token and self.current_token.type == TOKEN_NEWLINE:
                self.eat(TOKEN_NEWLINE)
        return prog_node

    def statement(self):
        if self.current_token.type == TOKEN_IDENTIFIER and self.peek() and self.peek().type == TOKEN_ASSIGN:
            return self.assignment_statement()
        else:
            return self.expression()

    def assignment_statement(self):
        var_name_token = self.eat(TOKEN_IDENTIFIER)
        self.eat(TOKEN_ASSIGN)
        value_node = self.expression()
        return VariableAssignNode(var_name_token, value_node)

    def atom(self):
        token = self.current_token

        if token.type == TOKEN_INTEGER:
            self.eat(TOKEN_INTEGER)
            return NumberNode(token)
        elif token.type == TOKEN_FLOAT:
            self.eat(TOKEN_FLOAT)
            return NumberNode(token)
        elif token.type == TOKEN_STRING:
            self.eat(TOKEN_STRING)
            return StringNode(token)
        elif token.type == TOKEN_IDENTIFIER:
            self.eat(TOKEN_IDENTIFIER)
            return VariableAccessNode(token)
        elif token.type == TOKEN_KEYWORD:
            keyword_val = token.value
            if keyword_val == 'Truth':
                self.eat(TOKEN_KEYWORD)
                return TruthNode(token)
            elif keyword_val == 'Falsity':
                self.eat(TOKEN_KEYWORD)
                return FalsityNode(token)
            elif keyword_val == 'Empty':
                self.eat(TOKEN_KEYWORD)
                return EmptyNode(token)
            elif keyword_val == 'display' or keyword_val == 'ask': # Added 'ask'
                self.eat(TOKEN_KEYWORD)
                return IdentifierNode(token) 
            else:
                self.error(message=f"Unexpected keyword '{keyword_val}' as an atom in an expression.")
        elif token.type == TOKEN_LPAREN:
            self.eat(TOKEN_LPAREN)
            node = self.expression()
            self.eat(TOKEN_RPAREN)
            return node
        
        self.error(message=f"Unexpected token in atom(): {token}")

    def call(self):
        node = self.atom()
        while self.current_token and self.current_token.type == TOKEN_LPAREN:
            if not isinstance(node, (IdentifierNode, VariableAccessNode)):
                 self.error(message="Expression is not callable.")

            self.eat(TOKEN_LPAREN)
            arg_nodes = []
            if self.current_token.type != TOKEN_RPAREN:
                arg_nodes.append(self.expression())
                while self.current_token.type == TOKEN_COMMA:
                    self.eat(TOKEN_COMMA)
                    arg_nodes.append(self.expression())
            self.eat(TOKEN_RPAREN)
            
            if isinstance(node, VariableAccessNode):
                func_name_node_for_call = IdentifierNode(node.var_name_token)
            elif isinstance(node, IdentifierNode):
                func_name_node_for_call = node
            else: 
                self.error(message="Internal parser error: Unexpected node type for function call.")
            node = FunctionCallNode(func_name_node_for_call, arg_nodes)
        return node

    def power(self):
        node = self.call()
        while self.current_token and self.current_token.type == TOKEN_POWER:
            op_token = self.eat(TOKEN_POWER)
            right_node = self.factor() 
            node = BinaryOpNode(left_node=node, op_token=op_token, right_node=right_node)
        return node

    def factor(self):
        token = self.current_token
        if token.type == TOKEN_PLUS:
            op_token = self.eat(TOKEN_PLUS)
            return UnaryOpNode(op_token, self.factor())
        elif token.type == TOKEN_MINUS:
            op_token = self.eat(TOKEN_MINUS)
            return UnaryOpNode(op_token, self.factor())
        return self.power()

    def term(self):
        node = self.factor()
        while self.current_token and self.current_token.type in (TOKEN_MULTIPLY, TOKEN_DIVIDE):
            op_token = self.current_token
            if op_token.type == TOKEN_MULTIPLY:
                self.eat(TOKEN_MULTIPLY)
            elif op_token.type == TOKEN_DIVIDE:
                self.eat(TOKEN_DIVIDE)
            node = BinaryOpNode(left_node=node, op_token=op_token, right_node=self.factor())
        return node

    def arith_expression(self):
        node = self.term()
        while self.current_token and self.current_token.type in (TOKEN_PLUS, TOKEN_MINUS):
            op_token = self.current_token
            if op_token.type == TOKEN_PLUS:
                self.eat(TOKEN_PLUS)
            elif op_token.type == TOKEN_MINUS:
                self.eat(TOKEN_MINUS)
            node = BinaryOpNode(left_node=node, op_token=op_token, right_node=self.term())
        return node
        
    def comp_expression(self):
        return self.arith_expression()

    def expression(self):
        return self.comp_expression()

    def parse(self):
        if not self.tokens or (len(self.tokens) == 1 and self.tokens[0].type == TOKEN_EOF):
            return ProgramNode()
        ast = self.program()
        if self.current_token and self.current_token.type != TOKEN_EOF:
            self.error(TOKEN_EOF, f"Expected EOF but found {self.current_token.type}")
        return ast


# rxscrite/nodes.py

class ASTNode:
    """Base class for all AST nodes."""
    def __init__(self, token=None): # Optional token for position tracking
        self.token = token
        self.line = token.line if token and hasattr(token, 'line') else None
        self.column = token.column if token and hasattr(token, 'column') else None

    def __repr__(self):
        return f"<{self.__class__.__name__}>"

# For expressions
class NumberNode(ASTNode):
    def __init__(self, token):
        super().__init__(token)
        self.value = token.value # The actual number (int or float)

    def __repr__(self):
        return f"NumberNode({self.value})"

class StringNode(ASTNode): # We'll use this later
    def __init__(self, token):
        super().__init__(token)
        self.value = token.value

    def __repr__(self):
        return f"StringNode({repr(self.value)})"

class IdentifierNode(ASTNode): # For variable names, function names
    def __init__(self, token):
        super().__init__(token)
        self.value = token.value # The name of the identifier

    def __repr__(self):
        return f"IdentifierNode({self.value})"

class UnaryOpNode(ASTNode):
    def __init__(self, op_token, node):
        super().__init__(op_token) # op_token is like TOKEN_MINUS for negation
        self.op_token = op_token
        self.node = node # The node to operate on (e.g., NumberNode)

    def __repr__(self):
        return f"UnaryOpNode({self.op_token}, {self.node})"

class BinaryOpNode(ASTNode):
    def __init__(self, left_node, op_token, right_node):
        super().__init__(op_token)
        self.left_node = left_node
        self.op_token = op_token
        self.right_node = right_node

    def __repr__(self):
        return f"BinaryOpNode({self.left_node}, {self.op_token}, {self.right_node})"

class VariableAssignNode(ASTNode):
    def __init__(self, var_name_token, value_node): # var_name_token is an IDENTIFIER token
        super().__init__(var_name_token)
        self.var_name_token = var_name_token
        self.value_node = value_node # Node representing the value being assigned

    def __repr__(self):
        return f"VariableAssignNode({self.var_name_token.value} = {self.value_node})"

class VariableAccessNode(ASTNode):
    def __init__(self, var_name_token): # var_name_token is an IDENTIFIER token
        super().__init__(var_name_token)
        self.var_name_token = var_name_token

    def __repr__(self):
        return f"VariableAccessNode({self.var_name_token.value})"

class FunctionCallNode(ASTNode):
    def __init__(self, func_name_node, arg_nodes): # func_name_node is often an IdentifierNode
        super().__init__(func_name_node.token if isinstance(func_name_node, ASTNode) else None)
        self.func_name_node = func_name_node
        self.arg_nodes = arg_nodes # List of nodes representing arguments

    def __repr__(self):
        return f"FunctionCallNode({self.func_name_node}, args={self.arg_nodes})"

class ProgramNode(ASTNode):
    """Represents a whole program or a block of statements."""
    def __init__(self):
        super().__init__() # No specific token for the whole program
        self.statements = [] # List of statement nodes

    def __repr__(self):
        return f"ProgramNode(statements_count={len(self.statements)})"

# --- Placeholder Nodes for Future Features ---
class IfNode(ASTNode): # 'check' statement
    def __init__(self, conditions_and_blocks, else_block):
        # conditions_and_blocks is a list of tuples: (condition_node, block_node)
        # else_block is a block_node or None
        # Attempt to get a representative token for line/col info
        rep_token = None
        if conditions_and_blocks and conditions_and_blocks[0] and conditions_and_blocks[0][0]:
             rep_token = conditions_and_blocks[0][0].token # Token from the first condition
        super().__init__(rep_token)
        self.conditions_and_blocks = conditions_and_blocks
        self.else_block = else_block

class ForLoopNode(ASTNode): # 'loop' statement
    def __init__(self, var_name_token, iterable_node, body_node, else_node):
        super().__init__(var_name_token)
        self.var_name_token = var_name_token
        self.iterable_node = iterable_node
        self.body_node = body_node
        self.else_node = else_node # Optional 'otherwise' block

class WhileLoopNode(ASTNode): # 'repeat_while' statement
    def __init__(self, condition_node, body_node, else_node):
        super().__init__(condition_node.token if condition_node else None) # Token from condition
        self.condition_node = condition_node
        self.body_node = body_node
        self.else_node = else_node # Optional 'otherwise' block

class FunctionDefNode(ASTNode): # 'func' statement
    def __init__(self, func_name_token, param_name_tokens, body_node, return_type_token=None):
        super().__init__(func_name_token)
        self.func_name_token = func_name_token
        self.param_name_tokens = param_name_tokens # List of IDENTIFIER tokens
        self.body_node = body_node
        # self.return_type_token = return_type_token # For optional type hinting later

# Boolean / Value Nodes
class TruthNode(ASTNode): # Represents the 'Truth' keyword
    def __init__(self, token):
        super().__init__(token)
        self.value = True
    def __repr__(self):
        return f"TruthNode({self.value})"

class FalsityNode(ASTNode): # Represents the 'Falsity' keyword
    def __init__(self, token):
        super().__init__(token)
        self.value = False
    def __repr__(self):
        return f"FalsityNode({self.value})"

class EmptyNode(ASTNode): # Represents the 'Empty' keyword
    def __init__(self, token):
        super().__init__(token)
        self.value = None # Or a special EmptyType object
    def __repr__(self):
        return f"EmptyNode({self.value})"

# You will add many more node types here as the language grows, e.g.:
# ListNode, DictNode, ReturnNode, ClassDefNode, ImportNode, etc.


# rxscrite/tokens.py

# Token types
# EOF (End Of File) token is used to indicate that
# there is no more input left for lexical analysis
TOKEN_EOF = 'EOF'

# Data types
TOKEN_INTEGER = 'INTEGER'        # e.g., 10, 234
TOKEN_FLOAT = 'FLOAT'          # e.g., 3.14, 0.5
TOKEN_STRING = 'STRING'        # e.g., "hello", 'world'
TOKEN_IDENTIFIER = 'IDENTIFIER'    # e.g., my_variable, my_function

# Keywords
TOKEN_KEYWORD = 'KEYWORD'        # e.g., display, check, loop (used for all keywords)

# Operators
TOKEN_PLUS = 'PLUS'            # +
TOKEN_MINUS = 'MINUS'           # -
TOKEN_MULTIPLY = 'MULTIPLY'        # *
TOKEN_DIVIDE = 'DIVIDE'           # /
TOKEN_POWER = 'POWER'           # ^

# Assignment
TOKEN_ASSIGN = 'ASSIGN'          # =

# Delimiters
TOKEN_LPAREN = 'LPAREN'        # (
TOKEN_RPAREN = 'RPAREN'        # )
TOKEN_COMMA = 'COMMA'           # , (for function arguments, list items etc.)
TOKEN_NEWLINE = 'NEWLINE'        # \n (important for statement separation)

# This dictionary maps keyword strings to their token type (TOKEN_KEYWORD)
# The lexer will use this to identify keywords.
RESERVED_KEYWORDS = {
    'display': TOKEN_KEYWORD,
    'Truth': TOKEN_KEYWORD,
    'Falsity': TOKEN_KEYWORD,
    'Empty': TOKEN_KEYWORD,
    'check': TOKEN_KEYWORD,
    'otherwise': TOKEN_KEYWORD,
    'orcheck': TOKEN_KEYWORD,
    'loop': TOKEN_KEYWORD,
    'repeat_while': TOKEN_KEYWORD,
    'func': TOKEN_KEYWORD,
    'blueprint': TOKEN_KEYWORD,
    'giveback': TOKEN_KEYWORD,
    'constructor': TOKEN_KEYWORD, # For classes
    'both': TOKEN_KEYWORD,    # Logical and
    'either': TOKEN_KEYWORD,  # Logical or
    'negate': TOKEN_KEYWORD,  # Logical not (can also be an operator depending on grammar)
    'inside': TOKEN_KEYWORD,  # Membership 'in'
    'skip': TOKEN_KEYWORD,    # pass
    'exit_loop': TOKEN_KEYWORD, # break
    'next_item': TOKEN_KEYWORD, # continue
    'use': TOKEN_KEYWORD,       # import
    'from_module': TOKEN_KEYWORD, # from
    'alias': TOKEN_KEYWORD,     # as
    'attempt': TOKEN_KEYWORD,   # try
    'on_error': TOKEN_KEYWORD,  # except
    'ensure': TOKEN_KEYWORD,    # finally
    'alert': TOKEN_KEYWORD,     # raise
    'verify': TOKEN_KEYWORD,    # assert
    'remove': TOKEN_KEYWORD,    # del
}


class Token:
    def __init__(self, type, value=None, line=None, column=None):
        self.type = type    # token type: INTEGER, PLUS, IDENTIFIER, etc.
        self.value = value  # token value: 10, "+", "my_var"
        self.line = line    # line number where token appears
        self.column = column # column number where token appears

    def __str__(self):
        """String representation of the class instance.

        Examples:
            Token(INTEGER, 10)
            Token(PLUS, '+')
            Token(IDENTIFIER, 'foo')
        """
        return f"Token({self.type}, {repr(self.value)}" + \
               (f", line={self.line}, col={self.column}" if self.line is not None else "") + ")"

    def __repr__(self):
        return self.__str__()

    def matches(self, type_, value_):
        """Checks if the token matches a given type and value."""
        return self.type == type_ and self.value == value_


# rxscrite/lexer.py

from .tokens import (
    Token,
    TOKEN_INTEGER, TOKEN_FLOAT, TOKEN_PLUS, TOKEN_MINUS, TOKEN_MULTIPLY,
    TOKEN_DIVIDE, TOKEN_POWER, TOKEN_LPAREN, TOKEN_RPAREN, TOKEN_EOF,
    TOKEN_IDENTIFIER, TOKEN_KEYWORD, RESERVED_KEYWORDS, TOKEN_ASSIGN,
    TOKEN_COMMA, TOKEN_NEWLINE, TOKEN_STRING # Added TOKEN_STRING
)
from .errors import RxLexerError

class Lexer:
    def __init__(self, text, filename="<stdin>"):
        self.text = text
        self.filename = filename
        self.pos = 0
        self.current_char = self.text[self.pos] if self.text else None
        self.line = 1
        self.column = 1

    def advance(self):
        """Advance the 'pos' pointer and set 'current_char'."""
        if self.current_char == '\n':
            self.line += 1
            self.column = 0 # Reset column, next advance will make it 1

        self.pos += 1
        if self.pos > len(self.text) - 1:
            self.current_char = None  # Indicates end of input
        else:
            self.current_char = self.text[self.pos]
        
        self.column +=1

    def peek(self):
        """Look at the next character without consuming the current one."""
        peek_pos = self.pos + 1
        if peek_pos > len(self.text) - 1:
            return None
        else:
            return self.text[peek_pos]

    def skip_whitespace(self):
        """Skips over whitespace characters."""
        while self.current_char is not None and self.current_char.isspace() and self.current_char != '\n':
            self.advance()

    def skip_comment(self):
        """Skips a single-line comment."""
        if self.current_char == '#':
            while self.current_char is not None and self.current_char != '\n':
                self.advance()

    def number(self):
        """Return a (multidigit) integer or float consumed from the input."""
        result = ''
        start_line = self.line
        start_column = self.column

        while self.current_char is not None and self.current_char.isdigit():
            result += self.current_char
            self.advance()

        if self.current_char == '.':
            result += '.'
            self.advance()
            if not (self.current_char and self.current_char.isdigit()):
                 self.error(f"Invalid character '{self.current_char if self.current_char else 'EOF'}' after '.' in number literal.")
            while self.current_char is not None and self.current_char.isdigit():
                result += self.current_char
                self.advance()
            return Token(TOKEN_FLOAT, float(result), start_line, start_column)
        else:
            return Token(TOKEN_INTEGER, int(result), start_line, start_column)

    def identifier(self):
        """Handle identifiers and reserved keywords."""
        result = ''
        start_line = self.line
        start_column = self.column

        if self.current_char is not None and (self.current_char.isalpha() or self.current_char == '_'):
            result += self.current_char
            self.advance()
        else: 
            self.error(f"Identifier cannot start with '{self.current_char}'")
            # This path should ideally not be reached if get_next_token calls this appropriately.
            # However, if it is, returning an EOF or a special error token might be needed,
            # but raising an error is generally better.
            return Token(TOKEN_EOF, None, self.line, self.column) # Fallback, should be unreachable

        while self.current_char is not None and (self.current_char.isalnum() or self.current_char == '_'):
            result += self.current_char
            self.advance()

        token_type = RESERVED_KEYWORDS.get(result, TOKEN_IDENTIFIER)
        if token_type == TOKEN_KEYWORD:
            return Token(token_type, result, start_line, start_column)
        return Token(token_type, result, start_line, start_column)

    def string_literal(self):
        """Handles string literals (e.g., "hello"). Does not support escape sequences yet."""
        result = ''
        start_line = self.line
        start_column = self.column
        quote_char = self.current_char # Should be " (or ' if you add support)

        self.advance() # Consume the opening quote

        while self.current_char is not None and self.current_char != quote_char:
            # TODO: Add escape sequence handling here (e.g., \n, \t, \", \\)
            # For now, all characters are taken literally.
            result += self.current_char
            self.advance()

        if self.current_char is None: # Unterminated string
            self.error("Unterminated string literal (missing closing quote).")
        
        self.advance() # Consume the closing quote
        return Token(TOKEN_STRING, result, start_line, start_column)

    def error(self, message=""):
        """Raises a lexer error."""
        full_message = f"Lexical error"
        if message:
            full_message += f": {message}"
        else:
            full_message += f" on '{self.current_char}'"
        raise RxLexerError(full_message, line=self.line, column=self.column)

    def get_next_token(self):
        """Lexical analyzer (also known as scanner or tokenizer)"""
        while self.current_char is not None:
            start_line = self.line
            start_column = self.column

            if self.current_char.isspace():
                if self.current_char == '\n':
                    self.advance()
                    return Token(TOKEN_NEWLINE, '\n', start_line, start_column)
                else:
                    self.skip_whitespace()
                    continue

            if self.current_char == '#':
                self.skip_comment()
                if self.current_char is None:
                    break 
                continue

            if self.current_char.isdigit():
                return self.number()

            if self.current_char.isalpha() or self.current_char == '_':
                return self.identifier()

            if self.current_char == '"': # Check for double-quoted strings
                return self.string_literal()
            
            # Add self.current_char == "'" for single-quoted strings if desired

            if self.current_char == '=':
                self.advance()
                return Token(TOKEN_ASSIGN, '=', start_line, start_column)

            if self.current_char == '+':
                self.advance()
                return Token(TOKEN_PLUS, '+', start_line, start_column)

            if self.current_char == '-':
                self.advance()
                return Token(TOKEN_MINUS, '-', start_line, start_column)

            if self.current_char == '*':
                self.advance()
                return Token(TOKEN_MULTIPLY, '*', start_line, start_column)

            if self.current_char == '/':
                self.advance()
                return Token(TOKEN_DIVIDE, '/', start_line, start_column)

            if self.current_char == '^':
                self.advance()
                return Token(TOKEN_POWER, '^', start_line, start_column)

            if self.current_char == '(':
                self.advance()
                return Token(TOKEN_LPAREN, '(', start_line, start_column)

            if self.current_char == ')':
                self.advance()
                return Token(TOKEN_RPAREN, ')', start_line, start_column)

            if self.current_char == ',':
                self.advance()
                return Token(TOKEN_COMMA, ',', start_line, start_column)
            
            char_val = self.current_char
            self.advance() # Consume the char to prevent infinite loop if error isn't raised or re-raised
            self.error(f"Unknown character: '{char_val}'")

        return Token(TOKEN_EOF, None, self.line, self.column)

    def tokenize(self):
        """Returns a list of all tokens from the input text."""
        tokens = []
        while True:
            token = self.get_next_token()
            tokens.append(token)
            if token.type == TOKEN_EOF:
                break
        return tokens


# rxscrite/interpreter.py

from .nodes import (
    NumberNode, UnaryOpNode, BinaryOpNode, VariableAssignNode,
    VariableAccessNode, FunctionCallNode, ProgramNode, IdentifierNode,
    StringNode, TruthNode, FalsityNode, EmptyNode
)
from .tokens import (
    TOKEN_PLUS, TOKEN_MINUS, TOKEN_MULTIPLY, TOKEN_DIVIDE, TOKEN_POWER
)
from .errors import RxRuntimeError, RxNameError, RxTypeError, RxZeroDivisionError

class Interpreter:
    def __init__(self, filename="<stdin>"):
        self.filename = filename
        self.symbol_table = {} 
        self._setup_builtins()
        self.current_node_line = None # For better error reporting in builtins

    def _setup_builtins(self):
        self.symbol_table['display'] = BuiltInFunction('display', self._builtin_display)
        self.symbol_table['ask'] = BuiltInFunction('ask', self._builtin_ask) # Added ask
        
        self.symbol_table['Truth'] = True
        self.symbol_table['Falsity'] = False
        self.symbol_table['Empty'] = None

    def _builtin_display(self, *args):
        output = " ".join(map(self._rx_str, args))
        print(output)
        return None

    def _builtin_ask(self, *prompt_args): # New builtin function
        prompt_message = ""
        if prompt_args:
            # Convert all prompt arguments to their RxScrite string form and join
            prompt_message = " ".join(map(self._rx_str, prompt_args))
        try:
            return input(prompt_message)
        except EOFError: # Handle cases where input stream might be closed (e.g. piping)
            return "Empty" # Or raise an RxScrite specific error

    def _rx_str(self, value):
        if value is True:
            return "Truth"
        if value is False:
            return "Falsity"
        if value is None:
            return "Empty"
        return str(value)

    def visit(self, node):
        # Store line number of current node for potential use in runtime errors from builtins
        if hasattr(node, 'line') and node.line is not None:
            self.current_node_line = node.line

        method_name = f'visit_{type(node).__name__}'
        visitor_method = getattr(self, method_name, self.generic_visit)
        return visitor_method(node)

    def generic_visit(self, node):
        raise RxRuntimeError(f"No visit_{type(node).__name__} method defined", line=node.line, column=node.column)

    def visit_ProgramNode(self, node):
        result = None
        for statement_node in node.statements:
            result = self.visit(statement_node)
        return result

    def visit_NumberNode(self, node):
        return node.value

    def visit_StringNode(self, node):
        return node.value

    def visit_TruthNode(self, node):
        return True

    def visit_FalsityNode(self, node):
        return False

    def visit_EmptyNode(self, node):
        return None

    def visit_UnaryOpNode(self, node):
        operand_value = self.visit(node.node)
        if not isinstance(operand_value, (int, float)):
            raise RxTypeError(f"Unary operator '{node.op_token.value}' not supported for type {type(operand_value).__name__}", line=node.line, column=node.column)
        if node.op_token.type == TOKEN_MINUS:
            return -operand_value
        elif node.op_token.type == TOKEN_PLUS:
            return +operand_value
        raise RxRuntimeError(f"Unknown unary operator: {node.op_token.type}", line=node.line, column=node.column)

    def visit_BinaryOpNode(self, node):
        left_val = self.visit(node.left_node)
        right_val = self.visit(node.right_node)

        # Handle string concatenation with '+'
        if node.op_token.type == TOKEN_PLUS:
            if isinstance(left_val, str) or isinstance(right_val, str):
                return self._rx_str(left_val) + self._rx_str(right_val)

        # Arithmetic operations for numbers
        if not (isinstance(left_val, (int, float)) and isinstance(right_val, (int, float))):
            err_msg = (f"Unsupported operand types for {node.op_token.value}: "
                       f"'{type(left_val).__name__}' and '{type(right_val).__name__}'")
            raise RxTypeError(err_msg, line=node.op_token.line, column=node.op_token.column)

        op_type = node.op_token.type
        try:
            if op_type == TOKEN_PLUS: # Should be numbers here due to above string check
                return left_val + right_val
            elif op_type == TOKEN_MINUS:
                return left_val - right_val
            elif op_type == TOKEN_MULTIPLY:
                return left_val * right_val
            elif op_type == TOKEN_DIVIDE:
                if right_val == 0:
                    raise RxZeroDivisionError("Division by zero", line=node.op_token.line, column=node.op_token.column)
                return left_val / right_val
            elif op_type == TOKEN_POWER:
                return left_val ** right_val
        except TypeError: # Should generally be caught by type checks above
            err_msg = (f"Problem with operand types for {node.op_token.value}: "
                       f"'{type(left_val).__name__}' and '{type(right_val).__name__}'")
            raise RxTypeError(err_msg, line=node.op_token.line, column=node.op_token.column)
        
        raise RxRuntimeError(f"Unknown binary operator: {op_type}", line=node.op_token.line, column=node.op_token.column)


    def visit_VariableAssignNode(self, node):
        var_name = node.var_name_token.value
        value = self.visit(node.value_node)
        self.symbol_table[var_name] = value
        return value

    def visit_VariableAccessNode(self, node):
        var_name = node.var_name_token.value
        # Keywords like Truth, Falsity, Empty are not accessed via VariableAccessNode
        # They are parsed into TruthNode, FalsityNode, EmptyNode directly.
        if var_name in self.symbol_table:
            return self.symbol_table[var_name]
        raise RxNameError(f"Name '{var_name}' is not defined", line=node.line, column=node.column)

    def visit_IdentifierNode(self, node):
        identifier_name = node.value
        # This node is used by FunctionCallNode for the function's name.
        # It could be a built-in like 'display' or 'ask', or a user-defined function later.
        # It can also be used by atom() for keywords like Truth, Falsity, Empty if they are directly used
        # in a way that results in an IdentifierNode (though parser now makes specific nodes for them).
        
        # For built-in functions/constants identified by keywords by the parser:
        if identifier_name in self.symbol_table:
            return self.symbol_table[identifier_name] # Returns BuiltInFunction or True/False/None
            
        raise RxNameError(f"Name '{identifier_name}' (from IdentifierNode) is not defined", line=node.line, column=node.column)


    def visit_FunctionCallNode(self, node):
        callable_obj = self.visit(node.func_name_node)

        if not isinstance(callable_obj, BuiltInFunction): # Later: add UserDefinedFunction
            func_name_str = node.func_name_node.value if hasattr(node.func_name_node, 'value') else 'unknown function'
            raise RxTypeError(f"'{func_name_str}' ({type(callable_obj).__name__}) is not a function", line=node.line, column=node.column)

        args = [self.visit(arg_node) for arg_node in node.arg_nodes]
        
        # Store current node's line for potential use in BuiltInFunction error reporting
        original_line = self.current_node_line
        if hasattr(node, 'line') and node.line is not None: # FunctionCallNode itself
            self.current_node_line = node.line

        try:
            return callable_obj.execute(args, self)
        except RxError: # Re-raise RxScrite errors directly
            raise
        except Exception as e: # Wrap other Python exceptions from builtins
            raise RxRuntimeError(f"Error during built-in function '{callable_obj.name}': {e}", line=self.current_node_line)
        finally:
            self.current_node_line = original_line


class BuiltInFunction:
    def __init__(self, name, python_callable=None):
        self.name = name
        self.python_callable = python_callable

    def execute(self, args, interpreter_instance):
        if self.python_callable:
            # The python_callable (e.g., _builtin_ask) will be called.
            # It can raise Python errors or RxErrors.
            return self.python_callable(*args)
        else:
            raise NotImplementedError(f"Built-in function '{self.name}' is not implemented.")

    def __repr__(self):
        return f"<BuiltInFunction:{self.name}>"